#!/usr/bin/env python
#
# PPU dump to nametable editor .sav converter
# Copyright 2012 Damian Yerrick
#
# Copying and distribution of this file, with or without
# modification, are permitted in any medium without royalty
# provided the copyright notice and this notice are preserved.
# This file is offered as-is, without any warranty.
#
"""

This program converts a PPU dump (16 KiB) to a SAV file for use with
the NES picture editor.

"""
from __future__ import with_statement
import sys


def parse_argv(argv):
    from optparse import OptionParser
    parser = OptionParser(usage="usage: %prog [options] [-i] INFILE [-o] OUTFILE")
    parser.add_option("-i", "--image", dest="infilename",
                      help="read 16 KiB PPU dump from INFILE", metavar="INFILE")
    parser.add_option("-o", "--output", dest="outfilename",
                      help="write SAV data to OUTFILE", metavar="OUTFILE")
    parser.add_option("-P", "--pattern-table", "--chr", dest="pattable",
                      help="read CHR data from ADDRESS ($0000 or $1000; default $0000)",
                      metavar="ADDRESS", default="0000")
    parser.add_option("-x", "--scroll-x", dest="xscroll",
                      help="scroll right by DISTANCE 16-pixel units (0-31, default 0)",
                      metavar="DISTANCE", type="int", default=0)
    parser.add_option("-y", "--scroll-y", dest="yscroll",
                      help="scroll down by DISTANCE 16-pixel units (0-29, default 0)",
                      metavar="DISTANCE", type="int", default=0)
    (options, args) = parser.parse_args(argv[1:])

    # Fill unfilled roles with positional arguments
    argsreader = iter(args)
    try:
        infilename = options.infilename
        if infilename is None:
            infilename = argsreader.next()
        outfilename = options.infilename
        if outfilename is None:
            outfilename = argsreader.next()
    except StopIteration:
        raise ValueError("not enough filenames")
    try:
        argsreader.next()
        raise ValueError("too many filenames")
    except StopIteration:
        pass

    pattable = options.pattable.lstrip('$')
    if pattable not in ('0000', '1000'):
        raise ValueError("pattern table must be $0000 or $1000")
    if options.yscroll < 0 or options.yscroll >= 30:
        raise ValueError("vertical scroll must be in 0-29")
    if options.xscroll < 0 or options.xscroll >= 32:
        raise ValueError("vertical scroll must be in 0-31")
    return (infilename, outfilename, int(pattable, 16), options.xscroll, options.yscroll)

def get_scrolled_nametable(four_screens, xscroll, yscroll):
    """Take a 1-screen slice from a 4-screen nametable, scrolled by 16*yscroll pixels."""

    from binascii import b2a_hex
    ntrows = [[screen[i:i + 32] for i in range(0, 960, 32)]
              for screen in four_screens]
    # reattach 4-screen
    ntrows = [l + r
              for (l, r) in zip(ntrows[0] + ntrows[2], ntrows[1] + ntrows[3])]
    # apply vertical scrolling
    toffset = yscroll * 2
    boffset = max(0, toffset - 30)
    ntrows = ntrows[toffset:toffset + 30] + ntrows[:boffset]
    # apply horizontal scrolling
    loffset = xscroll * 2
    roffset = max(0, loffset - 32)
    ntrows = [row[loffset:loffset + 32] + row[:roffset]
              for row in ntrows]
    
    # linearize attributes
    attrs = [[screen[i:i + 8] for i in range(960, 1024, 8)]
             for screen in four_screens]
    attrs = [l + r
             for (l, r) in zip(attrs[0] + attrs[2], attrs[1] + attrs[3])]
    attrs = [([((ord(c) >> 0) & 0x03, (ord(c) >> 2) & 0x03)
               for c in row],
              [((ord(c) >> 4) & 0x03, (ord(c) >> 6) & 0x03)
               for c in row])
             for row in attrs]
    attrs = [[subcol for byte in subrow for subcol in byte]
             for row in attrs for subrow in row]
    attrs = attrs[0:15] + attrs[16:31]

    # apply vertical scrolling
    boffset = max(0, yscroll - 15)
    attrs = attrs[yscroll:yscroll + 15] + attrs[:boffset]
    # apply horizontal scrolling
    roffset = max(0, xscroll - 16)
    attrs = [row[xscroll:xscroll + 16] + row[:roffset]
             for row in attrs]
    attrs.append([0]*16)

    # reassemble attributes
    attrs = [[row[i] | (row[i + 1] << 2) for i in range(0, 16, 2)]
             for row in attrs]
    attrs = [[tc | (bc << 4) for (tc, bc) in zip(t, b)]
             for (t, b) in zip(attrs[0::2], attrs[1::2])]
    attrs = "".join(chr(c) for row in attrs for c in row)

    ntrows.append(attrs)
##    print "\n".join(b2a_hex(row) for row in ntrows)
    return "".join(ntrows)

def main(argv=None):
    parsed = parse_argv(argv or sys.argv)
    (infilename, outfilename, pattable_base, xscroll, yscroll) = parsed
    with open(infilename, 'rb') as infp:
        data = infp.read(16384)
    bgchr = data[pattable_base:pattable_base + 0x1000]
    spritechr = data[0x0000:0x0800]
    nametables = [data[i + 0x2000:i + 0x2400] for i in range(0, 0x1000, 0x400)]
    nametable = get_scrolled_nametable(nametables, xscroll, yscroll)
    palette = data[0x3F00:0x3F20]
    with open(outfilename, 'wb') as outfp:
        outfp.writelines([bgchr, spritechr, nametable,
                          '\000'*0x300, palette, '\000'*0xE0])

if __name__=='__main__':
##    main(['ppu2sav.py', '-P1000', '-x5', '../docs/smb1 ppu image.bin', '../smb1.sav'])
    main()
